#include <SPI.h>
#include <LoRa.h>
#include <U8x8lib.h>
#include <WiFiClientSecure.h>

#define ss 18
#define rst 14
#define dio0 26

const char* ssid     = "Wifi-Network";          // WiFi SSID
const char* password = "PASSWORD";              // WiFi Password
const char* server = "testserver.aeq-web.com";  // Server URL
String gatewayid = "11AF12ADE32F";              // ID of Gateway

U8X8_SSD1306_128X64_NONAME_SW_I2C u8x8(15, 4, 16); //OLED LCD

String RxString, server_state;
int RxRSSI; boolean rx_fp;
long lastrx, crontimer, bctimer, wtimer;

WiFiClientSecure client;

void setup() {
  Serial.begin(9600);
  //-------------------Start OLED--------------------------
  u8x8.begin();
  u8x8.setFont(u8x8_font_chroma48medium8_r);
  u8x8.clearDisplay();
  u8x8.drawString(0, 1, "LoRa IoT Gateway");
  //-------------------Start OLED--------------------------

  //-------------------Start WiFi--------------------------
  u8x8.drawString(0, 3, "Connecting...");
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    Serial.print(".");
    delay(100);
  }
  u8x8.clearDisplay();
  u8x8.drawString(0, 1, "WiFi connected");
  server_state = "wait...";
  Serial.println();
  Serial.print("Connected to ");
  Serial.println(ssid);
  Serial.println("Starting LoRa...");
  //-------------------Start WiFi--------------------------

  //-------------------Start LoRA--------------------------
  SPI.begin(5, 19, 27, 18);
  LoRa.setPins(ss, rst, dio0);
  if (!LoRa.begin(868E6)) {
    Serial.println("Starting LoRa failed!");
    while (1);
  }
  Serial.println("-------LoRa initialized-------");
  //-------------------Start LoRA--------------------------
}

void loop() {
  //-------------------LoRa Receiver-----------------------
  int packetSize = LoRa.parsePacket();
  if (packetSize) {
    Serial.println("Receive LoRa Packet...");
    get_lora_data();
    call_https(RxString, RxRSSI, "gateway");
  }
  //-------------------LoRa Receiver-----------------------

  //-------------------Timer OLED Update-------------------
  if (millis() / 1000 > crontimer + 5) {
    crontimer = millis() / 1000;
    display_status();
  }
  //-------------------Timer OLED Update-------------------

  //-------------------Timer Send Beacon-------------------
  if (millis() / 1000 > bctimer + 300) {
    bctimer = millis() / 1000;
    call_https("beacon", 0, "beacon");
  }
  //-------------------Timer Send Beacon-------------------

  //-------------------Timer Check WiFi--------------------
  if (millis() / 1000 > wtimer + 600) {
    wtimer = millis() / 1000;
    check_wifi();
  }
  //-------------------Timer Check WiFi-------------------

  //-------------------Timer Reboot ESP-------------------
  if (millis() / 1000 > 90000) { //Reboot after a day
    ESP.restart();
  }
  //-------------------Timer Reboot ESP-------------------
}

void get_lora_data() {
  boolean clr = 1;
  rx_fp = true;
  while (LoRa.available()) {
    if (clr == 1) {
      RxString = "";
      clr = 0;
    }
    RxString += (char)LoRa.read();
  }
  RxRSSI = LoRa.packetRssi();
  Serial.print("Received Packet: ");
  Serial.println(RxString);
  lastrx = millis() / 1000;
  clr = 1;
}

void call_https(String iotdata, int iotrssi, String reqtype) {
  Serial.println("Connect to server via port 443");
  int wifirssi = WiFi.RSSI();
  if (!client.connect(server, 443)) {
    Serial.println("Connection failed!");
    server_state = "error";
  } else {
    Serial.println("Connected to server...");
    server_state = "online";
    client.print("POST ");
    client.print("/cn/gateway.php");
    client.println(" HTTP/1.1");
    client.print("Host: ");
    client.println(server);
    client.println("Content-Type: application/x-www-form-urlencoded");
    client.println("Connection: close");
    client.println("Content-Length: 128");
    client.println();
    client.print("data=");
    client.print(iotdata);
    client.print("&rssi=");
    client.print(iotrssi);
    client.print("&gwrssi=");
    client.print(wifirssi);
    client.print("&gwid=");
    client.print(gatewayid);
    client.print("&type=");
    client.print(reqtype);
    client.print("&end=end");
    client.println();
    delay(50);
    Serial.print("Waiting for response ");
    int res_timeout = 0;
    while (!client.available()) {
      delay(500);
      Serial.print(".");
      if (res_timeout++ >= 30) {
        client.stop();
        server_state = "timeout";
        Serial.println();
        Serial.println("Error: Connection aborted");
        break;
      }
    }

    Serial.println();
    Serial.println("-------Response Server-------");
    Serial.println();
    while (client.available()) {
      char c = client.read();
      Serial.write(c);
    }
    Serial.println();
    Serial.println("-------Response End-------");
    if (!client.connected()) {
      Serial.println("Server disconnected");
      client.stop();
    }
  }
}

void display_status() {
  u8x8.clearDisplay();
  long lastdata = ((millis() / 1000) - lastrx);
  u8x8.setCursor(0, 1);
  u8x8.print("Lastrx: ");
  if(rx_fp == true){
  u8x8.print(lastdata);
  u8x8.print("s");
  }else{
  u8x8.print("none");  
  }
  u8x8.setCursor(0, 3);
  u8x8.print("Server: ");
  u8x8.print(server_state);
  u8x8.setCursor(0, 7);
  u8x8.print("RSSI: ");
  u8x8.print(RxRSSI);
  u8x8.print(" dBm");
}

void check_wifi() {
  int wifi_retry;
  while (WiFi.status() != WL_CONNECTED && wifi_retry < 5 ) {
    wifi_retry++;
    Serial.println("WiFi not connected. Try to reconnect");
    WiFi.disconnect();
    WiFi.mode(WIFI_OFF);
    WiFi.mode(WIFI_STA);
    WiFi.begin(ssid, password);
    delay(5000);
  }
  if (wifi_retry >= 5) {
    Serial.println("Reboot");
    ESP.restart();
  }
}
